#!/usr/bin/python3

class Expression:
    def evaluate(self, scope):
        raise NotImplemented

class Number(Expression):
    def __init__(self, val):
        self.val = int(val)

    def evaluate(self, scope):
        return self.val

class Reference(Expression):
    def __init__(self, name):
        self.name = name

    def evaluate(self, scope):
        return scope[self.name]

class BinaryOperation(Expression):
    command = {
        '+': lambda l, r: l + r,
        '-': lambda l, r: l - r,
        '*': lambda l, r: l * r,
        '/': lambda l, r: l // r
    }

    def __init__(self, left, op, right):
        self.left = left
        self.op = op
        self.right = right

    def evaluate(self, scope):
        left = self.left.evaluate(scope)
        right = self.right.evaluate(scope)
        return self.command[self.op](left, right)

class UnaryOperation(Expression):
    command = {
        '+': lambda x: x,
        '-': lambda x: -x
    }

    def __init__(self, op, expr):
        self.op = op
        self.expr = expr

    def evaluate(self, scope):
        return self.command[self.op](self.expr.evaluate(scope))

if __name__ == '__main__':
    def tokenize(expr):
        def skip_ws(expr, pos):
            while pos != len(expr) and expr[pos].isspace():
                pos += 1
            return pos, None

        def number(expr, pos):
            token = ''
            while pos != len(expr) and expr[pos].isdigit():
                token += expr[pos]
                pos += 1
            return pos, token

        def name(expr, pos):
            if pos == len(expr) or not expr[pos].isalpha():
                return pos, None

            token = ''
            while pos != len(expr) and expr[pos].isalnum():
                token += expr[pos]
                pos += 1
            return pos, token

        def sign(expr, pos):
            if pos != len(expr) and expr[pos] in '+-*/()':
                return pos + 1, expr[pos]
            return pos, None

        parsers = (skip_ws, number, name, sign)
        pos = 0

        while pos != len(expr):
            old_pos = pos
            for parser in parsers:
                pos, token = parser(expr, pos)
                if token:
                    yield token
                    continue
            if pos == old_pos:
                raise Exception('Failed to parse at {}'.format(expr[pos:]))

    def parse(expr):
        def parse_binary(tokens, pos, min_prec = 1):
            if pos == len(tokens):
                raise Exception("Unexpected end of expression")

            precedence = { '+': 1, '-': 1, '*': 2, '/': 2}
            expr, pos = parse_prim(tokens, pos)
            current_prec = 0

            if pos == len(tokens):
                return expr, pos

            current_prec = precedence.get(tokens[pos], 0)
            while current_prec >= min_prec:
                operator = tokens[pos]
                while precedence.get(operator, 0) == current_prec:
                    right, pos = parse_binary(tokens, pos + 1, current_prec + 1)
                    expr = BinaryOperation(expr, operator, right)
                    if pos == len(tokens):
                        return expr, pos
                    operator = tokens[pos]
                current_prec -= 1

            return expr, pos

        def parse_prim(tokens, pos):
            if pos == len(tokens):
                raise Exception("Unexpected end of expression")

            if tokens[pos] == '(':
                expr, pos = parse_binary(tokens, pos + 1)
                if pos == len(tokens):
                    raise Exception('Unexpected end of expression')
                if tokens[pos] != ')':
                    raise Exception('Unexpected token {}'.format(tokens[pos]))
                return expr, pos + 1

            if tokens[pos] in ['+', '-']:
                op = tokens[pos]
                expr, pos = parse_prim(tokens, pos + 1)
                return UnaryOperation(op, expr), pos

            if tokens[pos].isdigit():
                return Number(tokens[pos]), pos + 1

            if tokens[pos].isalnum():
                return Reference(tokens[pos]), pos + 1

            raise Exception('Unexpected token {}'.format(tokens[pos]))

        return parse_binary([token for token in tokenize(expr)], 0)[0]

    assert parse('-2').evaluate({}) == -2
    assert parse('+2').evaluate({}) == 2
    assert parse('--2').evaluate({}) == 2
    assert parse('+-+-+--+++2').evaluate({}) == 2
    assert parse('a').evaluate({'a': 1}) == 1
    assert parse('-a').evaluate({'a': 1}) == -1
    assert parse('(a)').evaluate({'a': 1}) == 1
    assert parse('(((a)))').evaluate({'a': 1}) == 1
    assert parse('-(-(+a))').evaluate({'a': 1}) == 1
    assert parse('1+2').evaluate({}) == 3
    assert parse('1*2').evaluate({}) == 2
    assert parse('1+2+3+4+5+6').evaluate({}) == 21
    assert parse('1*2+3*4+5*6').evaluate({}) == 44
    assert parse('1+2*3+4*5+6').evaluate({}) == 33
    assert parse('(1+2)*(3+4)*(5+6)').evaluate({}) == 231
