#!/usr/bin/python3

import io

class CompressedStream(io.IOBase):
    def __init__(self, fileobj):
        self.fileobj = fileobj
        self.byte = b''
        self.count = 0

    def close(self):
        self.flush()
        self.fileobj.flush()

    def flush(self):
        self._write()
        self.fileobj.flush()

    def _write(self):
        if not self.count:
            return
        self.fileobj.write(bytes([self.count]) + self.byte)
        self.count = 0

    def write(self, bstr):
        for byte in bstr:
            if self.count == 255 or self.byte != byte:
                self._write()
            self.count += 1
            self.byte = bytes([byte])

    def _read(self, size=-1):
        if self.count:
            size = self.count if size < 0 or size >= self.count else size
            self.count -= size
            return self.byte * size

        data = self.fileobj.read(2)
        if not data:
            return b''

        self.count = data[0]
        self.byte = bytes([data[1]])
        return self._read(size)

    def read(self, size=-1):
        if size < 0:
            return b''.join(iter(lambda: self._read(), b''))

        data = b''
        while len(data) < size:
            bstr = self._read(size - len(data))
            if not bstr:
                break
            data += bstr
        return data

if __name__ == '__main__':
    from contextlib import closing

    data = bytes([0]) * 1024

    with closing(CompressedStream(open('file.bin', 'wb'))) as f:
        f.write(data)

    with closing(CompressedStream(open('file.bin', 'rb'))) as f:
        assert f.read() == data
